# FILE: plotconfig.tcl
#
# Tcl/Tk script for display configuration
#

###################### Plot Select and Configuration #####################
# 4-June-1998, mjd
#
# Local configuration results are stored in array psc.
Oc_Class Mmd_PlotSelectConfigure {
    const public variable winpath
    const public variable title

    # Import trace and initialization array.  Must be global.
    const public variable import_arrname = {}

    # Routine in mainline to call on an Apply or OK event.
    public variable apply_callback = Oc_Nop 

    # Menu and index (as a 2-item list) in the launching window
    # to toggle off in the constructor and on in the destructor.
    const public variable menu = {}

    # Internal state array
    private array variable psc

    # State (normal/disabled) dependency lists
    private array variable state_dep

    # Listbox update lists array
    private array variable listbox_dep

    Constructor {args} {
        # Create a (presumably) unique name for the toplevel
        # window based on the instance name.
        set winpath ".[$this WinBasename]"
	set title "Configure -- [Oc_Main GetInstanceName]"

	# Process user preferences
	eval $this Configure $args


        # Create dialog window
        toplevel $winpath
        wm group $winpath .
        wm title $winpath $title
        Ow_PositionChild $winpath ;# Position at +.25+/25 over '.'
        focus $winpath

        # Setup destroy binding.
	bind $winpath <Destroy> "$this WinDestroy %W"

        # Disable launch menu, if given
        if {![string match {} $menu]} {
	    set menuwin [lindex $menu 0]
	    if {[catch { $menuwin entryconfigure \
		    [lindex $menu 1] -state disabled } errmsg]} {
		Oc_Log Log $errmsg error
	    }
        }

        # Setup internal state array default values
        array set psc {
            colormaps  { Red-Black-Blue Blue-White-Red \
                    Black-Gray-White White-Green-Black }
	    quantitylist       {}
            arrow,status       1
            arrow,colormap     Red-Black-Blue
            arrow,colorcount   256
            arrow,quantity     z
            arrow,subsample    {}
            arrow,mag          1
            pixel,status       0
            pixel,colormap     Red-Black-Blue
            pixel,colorcount   256
            pixel,quantity     x
            pixel,subsample    0
            pixel,mag          1
            misc,drawboundary  1
            misc,background    white
            misc,margin        10
        }
        set gpsc [$this GlobalName psc]
        ## Global name of psc array.  This is needed to tie
        ## elements of this array to Tk buttons.

	set listbox_dep(colormaps) {}
	set listbox_dep(quantitylist) {}

        # Allow override from user initialization array
        if {![string match {} $import_arrname]} {
            upvar #0 $import_arrname userinit
            if {[info exists userinit]} {
                foreach elt [array names userinit] {
                    if {[info exists psc($elt)]} {
                        set psc($elt) $userinit($elt)
                    }
                }
            }
            trace variable userinit w "$this ExternalUpdate $import_arrname"
        }

        ########### Create dialog box
        # Grid Cells
        for {set i 0} {$i<3} {incr i} {
            for {set j 0} {$j<6} {incr j} {
                set f [frame $winpath.a$i$j -bd 2 -relief ridge]
                set a$i$j $f
                grid $f -row $i -column $j -sticky news
            }
        }

        # Column labels
        set clbl0 [label $a00.label -text "Plot Type"]
        set clbl1 [label $a01.label -text "Colormap"]
        set clbl2 [label $a02.label -text "# of Colors"]
        set clbl3 [label $a03.label -text "Color Quantity"]
        set clbl4 [label $a04.label -text "Subsample Rate"]
        set clbl5 [label $a05.label -text "Magnification"]

        pack $clbl0 $clbl1 $clbl2 $clbl3 $clbl4 $clbl5 -fill both

        # Arrow plot configuration
	set state_dep(arrow) {}
        set arrstat [checkbutton $a10.stat -text "Arrow" \
                -variable ${gpsc}(arrow,status) \
                -onvalue 1 -offvalue 0]
        $arrstat configure -command "$this UpdatePlotTypeState arrow"
        pack $arrstat -fill both

        set disabledcolor [$arrstat cget -disabledforeground] ;# Used
        ## below in Ow_EntryBox configuration

        Ow_ListBox New arrcm $a11.cm -height 3 \
                -variable ${gpsc}(arrow,colormap) \
                -outer_frame_options "-bd 0" \
		-listbox_options "-bd 2 -relief flat -width 0"
        eval $arrcm Insert end $psc(colormaps)
        pack $a11.cm -anchor e
        lappend state_dep(arrow) $arrcm
	lappend listbox_dep(colormaps) $arrcm

	Ow_EntryBox New arrcc $a12.cc -valuetype posint \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(arrow,colorcount)
        grid [$arrcc Cget -winpath]
        lappend state_dep(arrow) $arrcc

        Ow_ListBox New arrqt $a13.qt -height 3 \
                -variable ${gpsc}(arrow,quantity) \
                -outer_frame_options "-bd 0" \
		-listbox_options "-bd 2 -relief flat -width 0"
        eval $arrqt Insert end $psc(quantitylist)
        pack $a13.qt -fill both -expand 1
        lappend state_dep(arrow) $arrqt
	lappend listbox_dep(quantitylist) $arrqt

        Ow_EntryBox New arrsr $a14.sr -valuetype fixed \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(arrow,subsample)
        grid [$arrsr Cget -winpath]
        lappend state_dep(arrow) $arrsr

        Ow_EntryBox New arrmag $a15.mag -valuetype fixed \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(arrow,mag)
        grid [$arrmag Cget -winpath]
        lappend state_dep(arrow) $arrmag

        # Initialize arrow type display state
        $this UpdatePlotTypeState arrow

        # Pixel plot configuration
	set state_dep(pixel) {}
        set pixstat [checkbutton $a20.stat -text "Pixel" \
                -variable ${gpsc}(pixel,status) -onvalue 1 -offvalue 0]
        $pixstat configure -command "$this UpdatePlotTypeState pixel"
        pack $pixstat -fill both

        Ow_ListBox New pixcm $a21.cm -height 3 \
                -variable ${gpsc}(pixel,colormap) \
                -outer_frame_options "-bd 0" \
		-listbox_options "-bd 2 -relief flat -width 0"
        eval $pixcm Insert end $psc(colormaps)
        pack $a21.cm -anchor w
        lappend state_dep(pixel) $pixcm
	lappend listbox_dep(colormaps) $pixcm

	Ow_EntryBox New pixcc $a22.cc -valuetype posint \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(pixel,colorcount)
        grid [$pixcc Cget -winpath]
        lappend state_dep(pixel) $pixcc

        Ow_ListBox New pixqt $a23.qt -height 3 \
                -variable ${gpsc}(pixel,quantity) \
                -outer_frame_options "-bd 0" \
		-listbox_options "-bd 2 -relief flat -width 0"
        eval $pixqt Insert end $psc(quantitylist)
        pack $a23.qt -fill both -expand 1
        lappend state_dep(pixel) $pixqt
	lappend listbox_dep(quantitylist) $pixqt

        Ow_EntryBox New pixsr $a24.sr -valuetype fixed \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(pixel,subsample)
        grid [$pixsr Cget -winpath]
        lappend state_dep(pixel) $pixsr

        Ow_EntryBox New pixmag $a25.mag -valuetype fixed \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(pixel,mag)
        grid [$pixmag Cget -winpath]
        lappend state_dep(pixel) $pixmag
    
        # Initialize pixel type display state
        $this UpdatePlotTypeState pixel

        # Miscellaneous configuration
        set a30 [frame $winpath.a30 -bd 2 -relief ridge]
        grid $a30 -row 3 -column 0 -columnspan 6 -sticky news
        set f30 [frame $a30.f30 -bd 0]
        pack $f30 -side left -fill y
        
        set bdrycb [checkbutton $f30.bdrycb -text "Draw Boundary" \
                -variable ${gpsc}(misc,drawboundary) -onvalue 1 -offvalue 0]

        Ow_EntryBox New marginbox $f30.margin -valuetype fixed \
                -outer_frame_options "-bd 0 -relief flat" \
                -label "Margin:" -valuewidth 5 \
                -variable ${gpsc}(misc,margin)                

        grid $bdrycb x [$marginbox Cget -winpath] - -sticky w

        grid rowconfigure $winpath 3 -pad 10
        set bgclab [label $f30.bgclab -text "Background color:"]
        set bgc0 [radiobutton $f30.bgc0 -text "white" \
                -variable ${gpsc}(misc,background) -value "white"]
        set bgc1 [radiobutton $f30.bgc1 -text "black" \
                -variable ${gpsc}(misc,background) -value "black"]
        set bgc2 [radiobutton $f30.bgc2 -text "green" \
                -variable ${gpsc}(misc,background) -value "green"]
        grid $bgclab $bgc0 $bgc1 $bgc2 -sticky w

        # Control buttons
        set a40 [frame $winpath.a40 -bd 0]
        grid $a40 -row 4 -column 0 -columnspan 6 -sticky news
        grid rowconfigure $winpath 4 -pad 10
        set cancel [button $a40.cancel -text "Cancel" \
                -command "$this Action cancel" ]
        set apply [button $a40.apply -text "Apply" \
                -command "$this Action apply" ]
        set ok [button $a40.ok -text "OK" \
                -command "$this Action ok" ]
        pack $cancel $apply $ok -side left -expand 1
        wm protocol $winpath WM_DELETE_WINDOW "$cancel invoke"
	Ow_SetIcon $winpath
    }

    # Update from external source
    callback method ExternalUpdate { globalarrname name elt ops } {
        ## The trace is setup to pass in the global array name so that
        ## even if the trace is triggered by a call to Tcl_SetVar
        ## in C code called from a proc from which the traced array
        ## is not visible (due to no 'global' statement on that array),
        ## this trace can still function.        
        if {[string match {} $elt]} {return}
        upvar #0 $globalarrname source
        if {[info exists psc($elt)] && \
                ![string match $psc($elt) $source($elt)]} {
            set psc($elt) $source($elt)
	    # Reset listboxes as necessary
	    if {[info exists listbox_dep($elt)]} {
		foreach lb $listbox_dep($elt) {
		    $lb Remove 0 end
		    eval {$lb Insert end} $psc($elt)		    
		}
	    }
        }
    }

    # Plot type status change helper function
    callback method UpdatePlotTypeState { plottype } {
        if {$psc($plottype,status)} {
            set mystate normal
        } else {
            set mystate disabled
        }
        foreach win $state_dep($plottype) {
            $win configure -state $mystate
        }
    }

    # Dialog action routine
    callback method Action { cmd } {
        switch $cmd {
            cancel {
                $this Delete
                return
            }
            apply {
                eval $apply_callback psc
            }
            ok {
                eval $apply_callback psc
                $this Delete
                return
            }
            default {
                error "Illegal action code: $cmd"
            }
        }
    }

    # Destructor methods
    method WinDestroy { w } { winpath } {
	if {[string compare $winpath $w]!=0} { return } ;# Child destroy event
	$this Delete
    }
    Destructor {
        # Re-enable launching menu
        if {![string match {} $menu]} {
	    set menuwin [lindex $menu 0]
	    if {[winfo exists $menuwin]} {
		# If main window is going down, menu might already
		# be destroyed.
		if {[catch { $menuwin entryconfigure \
			[lindex $menu 1] -state normal } errmsg]} {
		    Oc_Log Log $errmsg error
		}
	    }
        }
        # Remove external update trace, if any
        if {[catch {
            if {![string match {} $import_arrname]} {
                upvar #0 $import_arrname userinit
                trace vdelete userinit w "$this ExternalUpdate $import_arrname"
            }
        } errmsg]} {
            Oc_Log Log $errmsg error
        }
        # Destroy windows
	if {[catch { 
            if {[info exists winpath] && [winfo exists $winpath]} {
                # Remove <Destroy> binding
                bind $winpath <Destroy> {}
                # Destroy toplevel, all children, and all bindings
                destroy $winpath
            }
        } errmsg]} {
            Oc_Log Log $errmsg error
        }
    }
}
